class CommandExecutionFailureException(Exception):
    pass


class RecordAlreadyExistsException(CommandExecutionFailureException):
    pass


class ObjectUpdateException(CommandExecutionFailureException):
    pass


class ObjectCreateException(CommandExecutionFailureException):
    pass


class ObjectDeletionException(CommandExecutionFailureException):
    pass


class FieldDoesNotExistException(Exception):
    pass


class QueryException(Exception):
    pass


class ImproperlyConfiguredFieldException(Exception):
    pass


class FieldValidationException(Exception):
    pass


class RecordUpdateException(Exception):
    pass


class CasFailureException(RecordUpdateException):
    pass



import json
import logging

import requests

from trood.api.custodian.command import Command
from trood.api.custodian.exceptions import CommandExecutionFailureException
from trood.api.custodian.records.manager import RecordsManager

handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)

logger = logging.getLogger(__name__)
logger.addHandler(handler)
logger.setLevel(logging.DEBUG)


class Client:
    __instances = {}
    _records_manager_class = RecordsManager

    server_url = None
    authorization_token = None

    def __init__(self, server_url: str, authorization_token: str = None):
        self.server_url = server_url.rstrip('/')
        self.authorization_token = authorization_token
        self.records = self._records_manager_class(self)

    def _make_query_string(self, params: dict):
        queries = []
        for key, value in params.items():
            queries.append('{}={}'.format(key, value))
        return '&'.join(queries)

    def execute(self, command: Command, data: dict = None, params: dict = None):
        """
        Performs call to the Custodian server API
        :param command:
        :param data:
        :param params:
        :return:
        :raises CommandExecutionFailureException:
        """
        url = '/'.join([self.server_url, command.name])
        logger.debug('Making {} request: url = "{}", json = "{}", query = "{}"'.format(
            command.method.upper(),
            url,
            json.dumps(data or {}),
            self._make_query_string(
                params or {}))
        )
        if self.authorization_token:
            headers = {'Authorization': self.authorization_token}
        else:
            headers = {}
        response = getattr(requests, command.method)(url, json=data, params=self._make_query_string(params or {}),
                                                     headers=headers)
        if response.content:
            response_content = response.json()
            if response_content['status'] == 'OK':
                return response_content.get('data', None), True
            else:
                return response_content.get('error', None), False
        else:
            if response.status_code == 204:
                return None, True
            else:
                raise CommandExecutionFailureException('Command execution failed')



import pytest
from hamcrest import assert_that, has_length, equal_to, not_, contains_string

from trood.api.custodian.client import Client


@pytest.mark.integration
def test_like_operator(client, two_records):
    """
    Filter records with "like" operator
    """
    fist_record_name_substring = 'odo'
    assert_that(two_records[0].name, contains_string(fist_record_name_substring))
    assert_that(two_records[1].name, not_(contains_string(fist_record_name_substring)))
    records = client.records.query('person').filter(name__like='%{}%'.format(fist_record_name_substring))
    assert_that(records, has_length(1))
    assert_that(records[0].get_pk(), equal_to(two_records[0].get_pk()))



class TestRqlSeries:
    @pytest.mark.integration
    def test_eq_operator(self, client, two_records):
        """
        Filter records with "eq" operator
        """
        fist_record_name = 'Feodor'
        assert_that(two_records[0].name, equal_to(fist_record_name))
        assert_that(two_records[1].name, not_(equal_to(fist_record_name)))
        records = client.records.query('person').filter(name__eq=fist_record_name)
        assert_that(records, has_length(1))
        assert_that(records[0].get_pk(), equal_to(two_records[0].get_pk()))

    @pytest.mark.integration
    def test_in_operator(self, client, two_records):
        """
        Filter records with "in" operator
        """
        second_record_name = 'Victor'
        assert_that(two_records[0].name, not_(equal_to(second_record_name)))
        assert_that(two_records[1].name, equal_to(second_record_name))
        records = client.records.query('person').filter(name__in=['Valentine', second_record_name])
        assert_that(records, has_length(1))
        assert_that(records[0].get_pk(), equal_to(two_records[1].get_pk()))

    @pytest.mark.integration
    def test_ne_operator(self, client, two_records):
        """
        Filter records with "ne" operator
        """
        fist_record_name = 'Feodor'
        assert_that(two_records[0].name, equal_to(fist_record_name))
        assert_that(two_records[1].name, not_(equal_to(fist_record_name)))
        records = client.records.query('person').filter(name__ne=fist_record_name)
        assert_that(records, has_length(1))
        assert_that(records[0].get_pk(), equal_to(two_records[1].get_pk()))

    @pytest.mark.integration
    def test_gt_operator(self, client, two_records):
        """
        Filter records with "gt" operator
        """
        age_value = 30
        assert_that(two_records[0].age < age_value)
        assert_that(two_records[1].age > age_value)
        records = client.records.query('person').filter(age__gt=age_value)
        assert_that(records, has_length(1))
        assert_that(records[0].get_pk(), equal_to(two_records[1].get_pk()))

    @pytest.mark.integration
    def test_ge_operator(self, client, two_records):
        """
        Filter records with "ge" operator
        """
        age_value = 20
        assert_that(two_records[0].age == age_value)
        assert_that(two_records[1].age > age_value)
        records = client.records.query('person').filter(age__ge=age_value)
        assert_that(records, has_length(2))

    @pytest.mark.integration
    def test_lt_operator(self, client, two_records):
        """
        Filter records with "lt" operator
        """
        age_value = 30
        assert_that(two_records[0].age < age_value)
        assert_that(two_records[1].age > age_value)
        records = client.records.query('person').filter(age__lt=age_value)
        assert_that(records, has_length(1))
        assert_that(records[0].get_pk(), equal_to(two_records[0].get_pk()))

    @pytest.mark.integration
    def test_le_operator(self, client, two_records):
        """
        Filter records with "le" operator
        """
        age_value = 20
        assert_that(two_records[0].age == age_value)
        assert_that(two_records[1].age > age_value)
        records = client.records.query('person').filter(age__le=age_value)
        assert_that(records, has_length(1))



from unittest import mock
import pytest

import django
from django.conf import settings
from django.db.models import Q, QuerySet
from django.db import connections, models
from django.test import TestCase
from pytest import mark
from rest_framework.exceptions import ValidationError, ErrorDetail
from rest_framework.test import APIRequestFactory
from rest_framework import exceptions

from trood.contrib.django.pagination import TroodRQLPagination
from trood.contrib.django.tests.settings import MockSettings

request_factory = APIRequestFactory()

if not settings.configured:
    settings.configure(default_settings=MockSettings)
from trood.contrib.django.filters import TroodRQLFilterBackend

django.setup()


class MockRelatedModel(models.Model):
    name = models.CharField(max_length=32)


class MockModel(models.Model):
    objects = None
    owner = models.ForeignKey(MockRelatedModel, on_delete=models.CASCADE)
    name = models.CharField(max_length=32)
    status = models.CharField(max_length=32)
    color = models.CharField(max_length=32)
    related = models.ManyToManyField(MockRelatedModel)


class ModelsFilteringTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        with connections['default'].schema_editor() as schema_editor:
            schema_editor.create_model(MockRelatedModel)
            schema_editor.create_model(MockModel)

    @classmethod
    def tearDownClass(cls):
        with connections['default'].schema_editor() as schema_editor:
            schema_editor.delete_model(MockModel)
            schema_editor.delete_model(MockRelatedModel)

    @mark.django_db
    def test_distinct_results(self):
        first = MockRelatedModel.objects.create(id=1, name="First")
        second = MockRelatedModel.objects.create(id=2, name="Second")
        third = MockRelatedModel.objects.create(id=3, name="Third")

        test_model = MockModel.objects.create(owner=first, name='Test', status='ACTIVE', color='RED')
        test_model.related.add(second, third)

        request = request_factory.get('/?rql=or(eq(owner.id,1),eq(related.id,1))')
        rows = TroodRQLFilterBackend().filter_queryset(request, MockModel.objects.all(), None)

        assert len(rows) == 1

    @mark.django_db
    def test_incorrect_rql_query(self):
        first = MockRelatedModel.objects.create(id=1, name="First")

        MockModel.objects.create(owner=first, name='Test', status='ACTIVE', color='RED')

        request = request_factory.get('/?rql=incorrect_rql(name,Test))')

        try:
            TroodRQLFilterBackend().filter_queryset(request, MockModel.objects.all(), None)
        except ValidationError as e:
            assert e.detail == [ErrorDetail(string='Incorrect rql query', code='invalid')]


def test_sort_parameter():
    rql = 'sort(-name,+id, test)'
    ordering = TroodRQLFilterBackend.get_ordering(rql)

    assert ordering == ['-name', 'id', 'test']


def test_like_filter():
    rql = 'like(name,"*23 test*")'
    filters = TroodRQLFilterBackend.parse_rql(rql)

    assert filters == [['like', 'name', '*23 test*']]

    queries = TroodRQLFilterBackend.make_query(filters)

    assert queries == [Q(('name__like', '*23 test*'))]

    assert str(
        MockModel.objects.filter(*queries).only('id', 'name').query
    ) == 'SELECT "tests_mockmodel"."id", "tests_mockmodel"."name" ' \
         'FROM "tests_mockmodel" ' \
         'WHERE "tests_mockmodel"."name" ' \
         'ILIKE %23 test% ESCAPE \'\\\''


def test_special_characters_1():
    rql = 'like(name,"test@test")'
    filters = TroodRQLFilterBackend.parse_rql(rql)

    assert filters == [['like', 'name', 'test@test']]

    queries = TroodRQLFilterBackend.make_query(filters)

    assert queries == [Q(('name__like', 'test@test'))]


def test_special_characters_2():
    rql = 'like(name,"test/test")'
    filters = TroodRQLFilterBackend.parse_rql(rql)

    assert filters == [['like', 'name', 'test/test']]

    queries = TroodRQLFilterBackend.make_query(filters)

    assert queries == [Q(('name__like', 'test/test'))]


def test_boolean_args():
    expected_true = [['exact', 'field', True]]

    assert TroodRQLFilterBackend.parse_rql('eq(field,True())') == expected_true
    assert TroodRQLFilterBackend.parse_rql('eq(field,true())') == expected_true
    assert TroodRQLFilterBackend.parse_rql('eq(field,True)') == expected_true
    assert TroodRQLFilterBackend.parse_rql('eq(field,true)') == expected_true

    expected_false = [['exact', 'field', False]]

    assert TroodRQLFilterBackend.parse_rql('eq(field,False())') == expected_false
    assert TroodRQLFilterBackend.parse_rql('eq(field,false())') == expected_false
    assert TroodRQLFilterBackend.parse_rql('eq(field,False)') == expected_false
    assert TroodRQLFilterBackend.parse_rql('eq(field,false)') == expected_false


def test_date_args():
    rql = "and(ge(created,2020-04-27T00:00:00.0+03:00),le(created,2020-05-03T23:59:59.9+03:00))"
    filters = TroodRQLFilterBackend.parse_rql(rql)

    assert filters == [
        ['AND', ['gte', 'created', '2020-04-27T00:00:00.0+03:00'], ['lte', 'created', '2020-05-03T23:59:59.9+03:00']]]


def test_default_grouping():
    rql = "eq(deleted,0),ge(created,2020-04-27T00:00:00.0+03:00),le(created,2020-05-03T23:59:59.9+03:00),sort(+id),limit(0,10)"

    filters = TroodRQLFilterBackend.parse_rql(rql)
    assert filters == [['AND', ['exact', 'deleted', '0'], ['gte', 'created', '2020-04-27T00:00:00.0+03:00'],
                        ['lte', 'created', '2020-05-03T23:59:59.9+03:00']]]


def test_mixed_grouping():
    rql = 'eq(deleted,0),or(eq(color,"red"),eq(status,2)),sort(+id),limit(0,10)'

    filters = TroodRQLFilterBackend.parse_rql(rql)
    assert filters == [['AND', ['exact', 'deleted', '0'], ['OR', ['exact', 'color', 'red'], ['exact', 'status', '2']]]]


@mock.patch.object(QuerySet, "count")
def test_rql_in_multiple_params(mocked_count):
    request = request_factory.get('/?rql=eq(status,1)&rql=in(color,(red,blue))&rql=limit(0,5)&rql=sort(id)')

    qs = TroodRQLFilterBackend().filter_queryset(request, MockModel.objects.all(), None)
    mocked_count.return_value = 10

    qs = TroodRQLPagination().paginate_queryset(qs.only('id'), request, None)

    assert str(
        qs.query
    ) == 'SELECT DISTINCT "tests_mockmodel"."id" ' \
         'FROM "tests_mockmodel" ' \
         'WHERE ("tests_mockmodel"."status" = 1 AND "tests_mockmodel"."color" IN (red, blue)) ' \
         'ORDER BY "tests_mockmodel"."id" ASC ' \
         'LIMIT 5'


def test_like_filter_case_insensitive():
    rql = 'like(name,"*23 TEST*")'
    filters = TroodRQLFilterBackend.parse_rql(rql)

    assert filters == [['like', 'name', '*23 TEST*']]

    queries = TroodRQLFilterBackend.make_query(filters)

    assert queries == [Q(('name__like', '*23 TEST*'))]
    assert str(MockModel.objects.filter(*queries).query) == 'SELECT "tests_mockmodel"."id", "tests_mockmodel"."owner_id", "tests_mockmodel"."name", "tests_mockmodel"."status", "tests_mockmodel"."color" FROM "tests_mockmodel" WHERE "tests_mockmodel"."name" ILIKE %23 TEST% ESCAPE \'\\\''


def test_not_eq_filter():
    rql = 'not(eq(name, "test"))'
    filters = TroodRQLFilterBackend.parse_rql(rql)
    queries = TroodRQLFilterBackend.make_query(filters)

    assert str(MockModel.objects.filter(*queries).query) == 'SELECT "tests_mockmodel"."id", "tests_mockmodel"."owner_id", "tests_mockmodel"."name", "tests_mockmodel"."status", "tests_mockmodel"."color" FROM "tests_mockmodel" WHERE NOT ("tests_mockmodel"."name" = test)'


def test_not_in_filter():
    rql = 'not(in(id,(1,2,3)))'
    filters = TroodRQLFilterBackend.parse_rql(rql)
    queries = TroodRQLFilterBackend.make_query(filters)
    assert str(MockModel.objects.filter(*queries).query) == 'SELECT "tests_mockmodel"."id", "tests_mockmodel"."owner_id", "tests_mockmodel"."name", "tests_mockmodel"."status", "tests_mockmodel"."color" FROM "tests_mockmodel" WHERE NOT ("tests_mockmodel"."id" IN (1, 2, 3))'


def test_not_or_filter():
    rql = 'not(or(in(id,(1,2,3)),eq(id, 123)))'
    filters = TroodRQLFilterBackend.parse_rql(rql)
    queries = TroodRQLFilterBackend.make_query(filters)
    assert str(MockModel.objects.filter(*queries).query) == 'SELECT "tests_mockmodel"."id", "tests_mockmodel"."owner_id", "tests_mockmodel"."name", "tests_mockmodel"."status", "tests_mockmodel"."color" FROM "tests_mockmodel" WHERE NOT (("tests_mockmodel"."id" IN (1, 2, 3) OR "tests_mockmodel"."id" = 123))'




from rest_framework.permissions import BasePermission


class TroodAuthPermissions(BasePermission):

    def has_permission(self, request, view):
        return True

    def has_object_permission(self, request, view, obj):
        return True


class TroodABACPermission(BasePermission):

    def has_permission(self, request, view):
        return request.abac.check_permited(request, view)

    def has_object_permission(self, request, view, obj):
        return request.abac.check_permited(request, view)



import requests
from django.conf import settings

from django.core.mail.message import EmailMessage
from django.core.mail.backends.smtp import EmailBackend
from trood.core.utils import get_service_token


class TroodEmailMessageTemplate(EmailMessage):
    def __init__(self, data=None, template=None, **kwargs):
        super().__init__(**kwargs)

        self.template = template
        self.data = data


class TroodEmailBackend(EmailBackend):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.mail_service_url = settings.MAIL_SERVICE_URL
        self.headers = {
            "Content-Type": "application/json", "Authorization": get_service_token()
        }

    def send_messages(self, email_messages):
        for message in email_messages:
            if isinstance(message, TroodEmailMessageTemplate):
                self._send_from_template(message)
            else:
                self._send_simple_mail(message)

    def _send_from_template(self, email_message: TroodEmailMessageTemplate):
        requests.post(
            f"{self.mail_service_url}/api/v1.0/mails/from_template/",
            headers=self.headers,
            json={
                # @todo: Remove hardcoded "mailbox" after implementing DEFAULT attribute for mailbox in TroodMail
                "mailbox": 1,
                "to": email_message.to,
                "bcc": email_message.bcc,
                "data": email_message.data,
                "template": email_message.template,
                "from_address": email_message.from_email
            }
        )

    def _send_simple_mail(self, email_message: EmailMessage):
        requests.post(
            f"{self.mail_service_url}/api/v1.0/mails/",
            headers=self.headers,
            json={
                # @todo: Remove hardcoded "mailbox" after implementing DEFAULT attribute for mailbox in TroodMail
                "mailbox": 1,
                "to": email_message.to,
                "bcc": email_message.bcc,
                "subject": email_message.subject,
                "body": email_message.body,
                "from_address": email_message.from_email
            }
        )



from hamcrest import *
import pytest

from trood.api.custodian.client import Client
from trood.api.custodian.records.model import Record
from hamcrest import assert_that, equal_to, contains_string, instance_of


@pytest.mark.integration
def test_client_updates_record_with_partial_data(client: Client):
    record = Record(obj='person', age=20, name='Ivan', is_active=True, street="Street")
    record = client.records.create(record)
    assert_that(record, instance_of(Record))

    # perform partial update
    record = client.records.partial_update('person', record.get_pk(), {'name': 'Petr'})
    assert_that(record.name, equal_to('Petr'))

    # check that new data got stored in Custodian
    record = client.records.get('person', record.get_pk())
    assert_that(record.name, equal_to('Petr'))



import json

import django
from django.conf import settings

from trood.contrib.django.auth.engine import TroodABACEngine
from trood.contrib.django.tests.settings import MockSettings

if not settings.configured:
    settings.configure(default_settings=MockSettings)
django.setup()

from django.http import HttpRequest
from rest_framework.response import Response
from trood.contrib.django.auth.renderers import TroodABACJSONRenderer
from rest_framework.viewsets import ViewSet


class MyView(ViewSet):
    renderer_classes = (TroodABACJSONRenderer, )
    action_map = {'get': 'list', 'post': 'create'}

    def get(self, request):
        return Response([{"field_to_hide": 100, "id": 1}])


def test_can_mask_simple_field():
    view = MyView()
    view.basename = 'TestView'
    view.action = 'list'
    request = HttpRequest()
    request.method = 'get'
    request.user = None

    setattr(request, 'abac', TroodABACEngine({"TEST": {"TestView": {"list": [{
        "result": "allow",
        "rule": {},
        "mask": ["field_to_hide"]
    }]}}}))
    request.data = None
    request.abac.check_permited(request, view)

    response = view.dispatch(request)
    response = view.get_renderers()[0].render(response.data, renderer_context={
        'request': request,
        'response': response,
    })
    assert 'field_to_hide' not in json.loads(response)[0]



import django
from django.conf import settings
from django.db.models import Model, IntegerField
from django.http import HttpRequest
from rest_framework.response import Response

from trood.contrib.django.auth.engine import TroodABACEngine
from trood.contrib.django.tests.settings import MockSettings


if not settings.configured:
    settings.configure(default_settings=MockSettings)
django.setup()

from rest_framework.views import APIView
from rest_framework.test import APIRequestFactory
from trood.contrib.django.auth.permissions import TroodABACPermission

request_factory = APIRequestFactory()


class MockModelOnlyID(Model):
    pass


class MyView(APIView):
    permission_classes = (TroodABACPermission, )

    def get(self, request):
        return Response({"id": 1}, 200)


def test_wildcard_rules():
    denied_view = MyView()
    denied_view.basename = 'DeniedView'
    denied_view.action = 'get'

    allowed_view = MyView()
    allowed_view.basename = 'AllowedView'
    allowed_view.action = 'get'

    request = HttpRequest()
    request.method = 'get'
    request.user = None

    setattr(request, 'abac', TroodABACEngine({
        "TEST": {
            "DeniedView": {"*": [{"result": "deny", "rule": {}, "mask": []}]},
            "AllowedView": {"*": [{"result": "allow", "rule": {}, "mask": []}]},
        }
    }))
    request.data = None

    response = denied_view.dispatch(request)
    assert response.status_code == 403

    response = allowed_view.dispatch(request)
    assert response.status_code == 200


def test_wildcard_sbj_id_rule_filter():
    view = MyView()
    view.basename = 'TestView'
    view.action = 'list'

    request = HttpRequest()
    request.method = 'get'
    request.user = {"id": 1}

    setattr(request, 'abac', TroodABACEngine({"TEST": {
        "_default_resolution": "deny",
        "TestView":
            {"*": [{
                "result": "allow",
                "rule": {"obj.id": "sbj.id"}
            }]}}}))
    request.data = None
    request.abac.check_permited(request, view)
    query = MockModelOnlyID.objects.filter(*request.abac.filters).query
    assert str(query) == 'SELECT "tests_mockmodelonlyid"."id" FROM "tests_mockmodelonlyid" WHERE "tests_mockmodelonlyid"."id" = 1'


import base64
import hashlib
import hmac
import os
from rest_framework.utils.serializer_helpers import ReturnDict


def get_attribute_path(obj, path: str, default=None):
    attributes = path.split(".")
    for i in attributes:
        if type(obj) is dict or type(obj) is ReturnDict:
            obj = obj.get(i, default)
        else:
            obj = getattr(obj, i, default)
    return obj


def get_service_token():
    domain = os.environ.get('SERVICE_DOMAIN')
    secret = os.environ.get('SERVICE_AUTH_SECRET')

    key = hashlib.sha1(b'trood.signer' + secret.encode('utf-8')).digest()
    signature = hmac.new(key, msg=domain.encode('utf-8'), digestmod=hashlib.sha1).digest()
    signature = base64.urlsafe_b64encode(signature).strip(b'=')
    return str(f'Service {domain}:{signature.decode("utf-8")}')



from django.core.management.color import no_style
from django.core.serializers.python import Deserializer
from django.db import transaction, connections, router, DatabaseError, IntegrityError, DEFAULT_DB_ALIAS
from rest_framework import viewsets, status
from rest_framework.response import Response


class TroodFixturesViewSet(viewsets.ViewSet):

    def create(self, request):
        objects_in_fixture = 0
        loaded_objects_count = 0
        models = set()

        fixture = request.data.get('fixture')

        if fixture:
            with transaction.atomic():
                connection = connections[DEFAULT_DB_ALIAS]

                with connection.constraint_checks_disabled():
                    objects = Deserializer(fixture, ignorenonexistent=True, handle_forward_references=True)
                    objs_with_deferred_fields = []

                    for obj in objects:
                        objects_in_fixture += 1
                        if router.allow_migrate_model(DEFAULT_DB_ALIAS, obj.object.__class__):
                            loaded_objects_count += 1
                            models.add(obj.object.__class__)
                            try:
                                obj.save()
                            # psycopg2 raises ValueError if data contains NUL chars.
                            except (DatabaseError, IntegrityError, ValueError) as e:
                                e.args = ("Could not load %(app_label)s.%(object_name)s(pk=%(pk)s): %(error_msg)s" % {
                                    'app_label': obj.object._meta.app_label,
                                    'object_name': obj.object._meta.object_name,
                                    'pk': obj.object.pk,
                                    'error_msg': e,
                                },)
                                raise
                        if obj.deferred_fields:
                            objs_with_deferred_fields.append(obj)

                    for obj in objs_with_deferred_fields:
                        obj.save_deferred_fields()

                if loaded_objects_count > 0:
                    sequence_sql = connection.ops.sequence_reset_sql(no_style(), models)
                    if sequence_sql:
                        with connection.cursor() as cursor:
                            for line in sequence_sql:
                                cursor.execute(line)
                # Close the DB connection -- unless we're still in a transaction. This
                # is required as a workaround for an edge case in MySQL: if the same
                # connection is used to create tables, load data, and query, the query
                # can return incorrect results. See Django #7572, MySQL #37735.
            if transaction.get_autocommit():
                connections[DEFAULT_DB_ALIAS].close()

        return Response({"objects_loaded": loaded_objects_count}, status=status.HTTP_200_OK)



import glob
import json
import importlib

from django.conf import settings
from django.apps import AppConfig


class TroodPluginsConfig(AppConfig):
    name = 'trood.contrib.django.apps.plugins'
    verbose_name = "Trood Plugins"

    @classmethod
    def ready(cls):
        try:
            from trood.contrib.django.apps.plugins.models import TroodPluginModel
            from trood.contrib.django.apps.plugins.core import TroodBasePlugin

            plugins = [
                module[:-3].replace("/", ".") for module in glob.glob(
                    '{}/*.py'.format(settings.TROOD_PLUGINS_PATH)
                )
            ]

            for plugin in plugins:
                importlib.import_module(plugin)

            for plugin in TroodBasePlugin.__subclasses__():
                obj, created = TroodPluginModel.objects.get_or_create(
                    pk=plugin.id, name=plugin.name, version=plugin.version
                )

                plugin_config = json.dumps(plugin.default_config)
                if created:
                    obj.config = plugin_config
                    obj.save()
                elif obj.config != plugin_config:
                    obj.config = plugin_config
                    obj.save()

                plugin.register()
        except Exception as e:
            print("Can't load plugins, exception: {}".format(e))




import os

import pytest

from trood.api.custodian.client import Client
from trood.api.custodian.records.model import Record


@pytest.fixture
def client():
    return Client(server_url=os.environ.get('SERVER_URL', 'http://localhost:8000/custodian'))


@pytest.fixture
def person_record():
    return Record(obj='person', age=20, name='Ivan', is_active=True, street="Street")



@pytest.fixture
def two_records(client):
    client.records.delete(*[x for x in client.records.query('person')])
    first_record = Record(obj='person',
                          **{'name': 'Feodor', 'is_active': True, 'age': 20, 'street': 'street'})
    second_record = Record(obj='person',
                           **{'name': 'Victor', 'is_active': False, 'age': 40, 'street': 'street'})
    first_record, second_record = client.records.create(first_record, second_record)
    return first_record, second_record




from copy import deepcopy
from trood.api.custodian.exceptions import QueryException


class Q:
    _query = None
    _logical_expressions = None
    _inverted = None

    _KNOWN_OPERATORS = ('in', 'like', 'eq', 'ne', 'gt', 'ge', 'lt', 'le', 'is_null')

    def __init__(self, **kwargs):
        self._query = deepcopy(kwargs)
        self._logical_expressions = []
        self._inverted = False

    def __and__(self, other):
        self._logical_expressions.append({
            'operator': 'and',
            'query': other
        })
        return self

    def __or__(self, other):
        self._logical_expressions.append({
            'operator': 'or',
            'query': other
        })
        return self

    def __invert__(self):
        self._inverted = not self._inverted
        return self

    def to_string(self):
        """
        Constructs a string representation of RQL query
        :return:
        """
        # construct query first
        expressions = []
        for key, value in self._query.items():
            operator, field = self._parse_key(key)
            value = self._normalize_value(value)
            expressions.append('{}({},{})'.format(operator, field, value))
        query_string = ','.join(expressions)
        # and then apply logical operators
        for logical_expression in self._logical_expressions:
            query_string = '{}({},{})'.format(
                logical_expression['operator'], query_string, logical_expression['query'].to_string()
            )
        # apply inversion operator
        if self._inverted:
            query_string = 'not({})'.format(query_string)
        return query_string

    def _parse_key(self, key):
        """
        Parses the key to a field and an operator.
        Example: person__age__gt will be parsed to the "person.age" field and the "gt" operator
        :param key:
        :return:
        """
        split_key = key.split('__')
        operator = split_key[-1]
        field = '.'.join(split_key[:-1])
        if operator not in self._KNOWN_OPERATORS:
            raise QueryException('"{}" operator is unknown'.format(operator))
        return operator, field

    def _normalize_value(self, value):
        """
        Returns RQL-friendly value
        :param value:
        :return:
        """
        if isinstance(value, (list, tuple)):
            return '({})'.format(','.join([str(x) for x in value]))
        return value


def evaluate(func):
    def wrapper(*args, **kwargs):
        self = args[0]
        if not self._is_evaluated:
            self._evaluate()
        return func(*args, **kwargs)

    return wrapper


class Query:
    _q_objects = None
    _orderings = None
    _limit = None
    _is_evaluated = None
    _result = None
    _depth = 1
    _omit_outers = False

    def __init__(self, obj_name: str, manager, depth=1, omit_outers=False):
        self._obj_name = obj_name
        self._manager = manager
        self._q_objects = []
        self._orderings = []
        self._limit = None
        self._is_evaluated = False
        self._result = None
        self._depth = depth
        self._omit_outers = omit_outers

    def filter(self, q_object: Q = None, **filters):
        """
        Applies filters to the current query
        :param q_object:
        :param filters:
        :return:
        """
        new_query = QueryFactory.clone(self)
        if q_object:
            new_query._q_objects.append(q_object)
        if filters:
            new_query._q_objects.append(Q(**filters))
        return new_query

    def to_string(self) -> str:
        """
        Assembles query into a RQL expression, multiple Q objects will be interpreted as an "AND" expression`s
        components
        :return:
        """
        # queries
        if self._q_objects:
            query_string = self._q_objects[0].to_string()
            for q_object in self._q_objects[1:]:
                query_string = '{}({},{})'.format(
                    'and', query_string, q_object.to_string()
                )
        else:
            query_string = ''
        # ordering options
        if self._orderings:
            ordering_expression = 'sort({})'.format(', '.join(self._orderings))
            query_string = ','.join(filter(lambda x: bool(x), [query_string, ordering_expression]))
        # limit option
        if self._limit:
            limit_expression = 'limit({},{})'.format(*self._limit)
            query_string = ','.join(filter(lambda x: bool(x), [query_string, limit_expression]))
        return query_string

    def order_by(self, *orderings: str):
        """
        Sets ordering to the Query object.
        Examples:
        query.order_by('person__last_name')
        query.order_by('-person__last_name') #descending ordering
        :param ordering:
        :return:
        """
        new_query = QueryFactory.clone(self)
        for ordering in orderings:
            ordering = ordering.replace('__', '.')
            new_query._orderings.append(ordering)
        return new_query

    def __getitem__(self, item):
        """
        If slice is used __getitem__ returns new query with offset-limit applied,
        otherwise returns item by index
        :param item:
        :return:
        """
        if self._limit:
            raise Exception('Cannot limit already limited query')
        new_query = QueryFactory.clone(self)
        if isinstance(item, slice):
            offset = item.start or 0
            limit = item.stop - offset
            new_query._limit = (offset, limit)
            return new_query
        else:
            if not self._is_evaluated:
                self._evaluate()
            return self._result[item]

    @evaluate
    def __iter__(self):
        return self._result.__iter__()

    @evaluate
    def __len__(self):
        return self._result.__len__()

    def _evaluate(self):
        """
        Evaluates the query using RecordsManager
        """
        records = self._manager._query(self._obj_name, self.to_string(), depth=self._depth, omit_outers=self._omit_outers)
        self._result = records
        self._is_evaluated = True
        return self._result


class QueryFactory:
    @classmethod
    def clone(cls, query: Query):
        new_query = Query(obj_name=query._obj_name, manager=query._manager)
        new_query._q_objects = query._q_objects[:]
        new_query._orderings = query._orderings[:]
        new_query._limit = query._limit
        new_query._depth = query._depth
        new_query._omit_outers = query._omit_outers
        return new_query



from rest_framework import filters


class TroodABACFilterBackend(filters.BaseFilterBackend):

    def filter_queryset(self, request, queryset, view):
        return request.abac.filter_data(queryset)



import json

import requests
import os

from django.contrib.auth.models import AnonymousUser
from requests import HTTPError
from rest_framework.authentication import BaseAuthentication, get_authorization_header
from rest_framework import exceptions
from django_redis import get_redis_connection

from trood.core.utils import get_service_token


class TroodUser(object):
    def __init__(self, object):
        for k, v in object.items():
            self.__setattr__(k, v)

    @property
    def is_authenticated(self):
        return True


class TroodTokenAuthentication(BaseAuthentication):

    def authenticate(self, request):
        auth = get_authorization_header(request).decode('utf-8')
        parts = auth.split()

        if not auth or len(parts) != 2:
            return AnonymousUser(), None

        try:
            user = self.get_user_from_cache(auth)

            if not user:
                user = self.get_user_from_service(auth)

            return user, parts[1]

        except Exception:
            raise exceptions.AuthenticationFailed()

    def get_user_from_cache(self, token):
        cache_type = os.environ.get("CACHE_TYPE", None)

        if cache_type == "REDIS":
            redis = get_redis_connection("default")
            cached_user = redis.get(f"AUTH:{token}")

            if cached_user:
                return TroodUser(json.loads(cached_user))

        return None

    def get_user_from_service(self, token):
        parts = token.split()

        if not parts or len(parts) != 2:
            return None

        try:
            token_type = "service" if parts[0] == "Service" else "user"

            response = requests.post(
                "{}api/v1.0/verify-token/".format(os.environ.get('TROOD_AUTH_SERVICE_URL')),
                data={
                    "type": token_type,
                    "token": parts[1]
                },
                headers={"Authorization": get_service_token()},
            )

            response.raise_for_status()
            response_decoded = response.json()
            return TroodUser(response_decoded['data'])

        except HTTPError:
            raise exceptions.AuthenticationFailed()



from rest_framework import viewsets
from rest_framework.permissions import IsAuthenticated

from trood.contrib.django.apps.plugins.models import TroodPluginModel
from trood.contrib.django.apps.plugins.serialiers import TroodPluginSerizalizer


class TroodPluginsViewSet(viewsets.ModelViewSet):
    queryset = TroodPluginModel.objects.all()
    permission_classes = (IsAuthenticated, )
    serializer_class = TroodPluginSerizalizer



import os

class MockSettings:
    LANGUAGE_CODE = 'en-us'
    ALLOWED_HOSTS = ['*']
    EMAIL_BACKEND = ''
    MIGRATION_MODULES = []
    DEFAULT_INDEX_TABLESPACE = ''
    CACHES = ''
    REST_FRAMEWORK = {
        'DEFAULT_AUTHENTICATION_CLASSES': [],
        'DEFAULT_PERMISSION_CLASSES': [],
        'UNAUTHENTICATED_USER': None,
    }
    INSTALLED_APPS = [
        'trood.contrib.django.tests',
    ]
    DEBUG = True
    LOGGING_CONFIG = {}
    LOGGING = {}
    SECRET_KEY = ''
    FORCE_SCRIPT_NAME = ''
    DEFAULT_TABLESPACE = ''
    DEFAULT_CHARSET = 'utf-8'
    DATABASE_ROUTERS = []
    DATA_UPLOAD_MAX_MEMORY_SIZE = 2621440
    DATA_UPLOAD_MAX_NUMBER_FIELDS = 1000
    BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    DATABASES = {'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3')
    }}
    ABSOLUTE_URL_OVERRIDES = {}
    SERVICE_DOMAIN = "TEST"
    USE_I18N = False



from rest_framework.renderers import JSONRenderer


def render_callback(renderer):
    """Wrapper which call base class render if no data returned"""
    def wrapped_renderer(*args, **kwargs):
        response = renderer(*args, **kwargs)
        if not response:
            response = JSONRenderer.render(*args, **kwargs)
        return response
    return wrapped_renderer


class TroodABACJSONRenderer(JSONRenderer):
    """Renderer which mask data by abac rule and pass through"""
    @render_callback
    def render(self, data, accepted_media_type=None, renderer_context=None):
        renderer_context = renderer_context or {}
        if not ('request' in renderer_context and hasattr(renderer_context['request'], 'abac')):
            return
        abac = renderer_context['request'].abac
        for defer in abac.mask:
            if isinstance(data, list):
                for d in data:
                    d.pop(defer)
            if isinstance(data, dict):
                data.pop(defer)
        return super().render(data, accepted_media_type=accepted_media_type, renderer_context=renderer_context)



import re

from rest_framework.response import Response
from rest_framework.pagination import BasePagination
from rest_framework.settings import api_settings
from django.conf import settings


class TroodRQLPagination(BasePagination):
    default_limit = api_settings.PAGE_SIZE

    query_param = getattr(settings, 'RQL_FILTER_QUERY_PARAM', 'rql')

    def get_paginated_response(self, data):
        return Response({'data': data, 'total_count': self.count})

    def paginate_queryset(self, queryset, request, view=None):
        self.count = self.get_count(queryset)
        offset, limit = self.get_limit_offset(request)

        if limit is None:
            return None

        if self.count == 0 or offset > self.count:
            return []
        return queryset[offset:offset + limit]

    def get_count(self, queryset):
        """
        Determine an object count, supporting either querysets or regular lists.
        """
        try:
            return queryset.count()
        except (AttributeError, TypeError):
            return len(queryset)

    def get_limit_offset(self, request):
        if self.query_param in request.GET:
            limit_parsed = re.search(
                'limit\((?P<offset>\d+),\s*(?P<limit>\d+)\)', ','.join(
                    request.GET.getlist(self.query_param, [])
                )
            )

            if limit_parsed:
                groups = limit_parsed.groups()

                if len(groups) == 2:
                    return int(groups[0]), int(groups[1])

        return 0, self.default_limit



import pytest
from hamcrest import *

from trood.api.custodian.client import Client
from trood.api.custodian.exceptions import QueryException
from trood.api.custodian.records.query import Q, Query


def test_q():
    """
    Tests regular Q-expression`s string representation
    """
    queryset = (Q(age__gt=18) | Q(age__lt=53)) & Q(is_active__eq=True)
    assert_that(queryset.to_string(), equal_to('and(or(gt(age,18),lt(age,53)),eq(is_active,True))'))


def test_q_with_list_value():
    """
    Tests regular Q-expression`s string representation with a list as a value
    """
    queryset = Q(city_id__in=[1, 4, 7])
    assert_that(queryset.to_string(), equal_to('in(city_id,(1,4,7))'))


def test_q_unknown_operator_raises_exception():
    """
    Tests regular Q-expression`s string representation
    """
    queryset = (Q(age__gt=18) | Q(age__lt=53)) & Q(is_active__custom_operator=True)
    with pytest.raises(QueryException):
        queryset.to_string()


def test_inverted_q():
    """
    Tests Q-expression`s string representation when using the "~" operator
    """
    queryset = (Q(age__gt=18) | Q(age__lt=53)) & ~Q(is_active__eq=True)
    assert_that(queryset.to_string(), equal_to('and(or(gt(age,18),lt(age,53)),not(eq(is_active,True)))'))


def test_query():
    query = Query('person', None).filter((Q(age__gt=18) | Q(age__lt=53)) & Q(is_active__eq=True)) \
        .filter(address__city__name__eq='St. Petersburg')
    assert_that(
        query.to_string(),
        equal_to('and(and(or(gt(age,18),lt(age,53)),eq(is_active,True)),eq(address.city.name,St. Petersburg))')
    )


def test_query_ordering():
    query = Query('person', None).filter(is_active__eq=True).order_by('+person__last_name',
                                                                           '-person__phone_number')
    assert_that(query.to_string(), contains_string('sort(+person.last_name, -person.phone_number)'))


def test_query_slicing():
    query = Query('person', None).filter(is_active__eq=True)[150:200]
    assert_that(query.to_string(), contains_string('limit(150,50)'))


def test_query_access_by_index():
    query = Query('person', None).filter(is_active__eq=True)
    query._is_evaluated = True
    query._result = [1, 2, 3, 4, 5]
    assert_that(query[3], equal_to(4))


def test_query_resets_evaluated_result_on_query_modifications():
    client = Client(server_url='http://mocked/custodian')
    query = Query('person', client.records).filter(is_active__eq=True)
    query._is_evaluated = True
    assert_that(query._is_evaluated)
    query = query.filter(client__first_name__eq='Ivan')
    assert_that(not query._is_evaluated)


def test_query_filter_operation_does_not_affect_existing_query():
    base_query = Query('person', None).filter(Q(is_active__eq=True))
    updated_query = base_query.filter(address__city__name__eq='St. Petersburg')
    assert_that(base_query.to_string(), equal_to('eq(is_active,True)'))
    assert_that(updated_query.to_string(), equal_to('and(eq(is_active,True),eq(address.city.name,St. Petersburg))'))


def test_empty_query_with_limits_assembles_correct_expression():
    base_query = Query('person', None)[:100]
    assert_that(base_query.to_string(), equal_to('limit(0,100)'))



class COMMAND_METHOD:
    DELETE = 'delete'
    GET = 'get'
    POST = 'post'
    PATCH = 'patch'

    @classmethod
    def get_available_methods(cls):
        return [value for key, value in cls.__dict__.items() if not key.startswith('__') and isinstance(value, str)]


class Command:
    name: str
    method: str

    def __init__(self, name: str, method: str):
        assert method in COMMAND_METHOD.get_available_methods(), 'Method must be one of the COMMAND_METHOD`s attribute'
        self.name = name
        self.method = method

    def __repr__(self):
        return '<Command name="{}" method="{}">'.format(self.name, self.method)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2018-12-12 16:36
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='TroodPluginModel',
            fields=[
                ('id', models.CharField(max_length=128, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=128)),
                ('version', models.CharField(max_length=32)),
                ('active', models.BooleanField(default=False)),
                ('config', models.TextField()),
            ],
        ),
    ]



from rest_framework import serializers


class TroodDynamicSerializer(serializers.ModelSerializer):
    def __init__(self, *args, **kwargs):
        super(TroodDynamicSerializer, self).__init__(*args, **kwargs)
        request = self.context.get('request', None)
        if self.fields and request:
            exclude = request.GET.get('exclude', None)
            only = request.GET.get('only', None)
            if exclude:
                for a in exclude.split(','):
                    self.fields.pop(a)
            elif only:
                only = only.split(',')
                keys = list(self.fields.keys())
                for a in keys:
                    if a not in only:
                        self.fields.pop(a)



from operator import __or__
from functools import reduce

from django.conf import settings
from django.db.models import Q
from rest_framework.exceptions import PermissionDenied

from trood.core.utils import get_attribute_path


class TroodABACResource(object):

    def __init__(self, name=None):
        self.name = name

    def __call__(self,  obj):
        setattr(obj, "_is_abac", True)

        if self.name:
            setattr(obj, "_abac_resource", self.name)
        else:
            setattr(obj, "_abac_resource", obj.__name__)

        return obj


class TroodABACEngine:
    filters = []
    mask = []

    def __init__(self, rules=None):
        self.rules = rules

    def check_permited(self, request, view):
        paths = [f"{view.basename}.{view.action}", f"{view.basename}.*", f"*.{view.action}", "*.*"]
        rules = []
        for p in paths:
            rules.extend(
                get_attribute_path(self.rules, settings.SERVICE_DOMAIN + "." + p, default=[])
            )

        resolver = TroodABACResolver(subject=request.user, context=request.data)

        for rule in rules:
            passed, result, filters = resolver.evaluate_rule(rule)
            if passed and result == 'allow':
                self.filters = filters
                self.mask = rule.get('mask')
                return True
            elif passed and result == 'deny':
                raise PermissionDenied("Access restricted by ABAC access rule")

        if get_attribute_path(self.rules, settings.SERVICE_DOMAIN + "._default_resolution") == 'allow' \
                or get_attribute_path(self.rules, "_default_resolution") == 'allow':
            return True

        raise PermissionDenied("Access restricted by ABAC access rule")

    def filter_data(self, data):
        return data.filter(*self.filters)


class TroodABACResolver:

    def __init__(self, subject={}, context={}):
        self.data_source = {
            "sbj": subject,
            "ctx": context
        }

    def evaluate_rule(self, rule: dict) -> (bool, str, list):
        passed, filters = self.evaluate_condition(rule['rule'])

        return passed, rule['result'], filters

    def evaluate_condition(self, condition):
        filters = []
        result = True
        operator = "exact"

        for operand, value in condition.items():
            if type(value) is dict:
                operator, value = list(value.items())[0]

            elif type(value) is list:
                operator = operand

            operand, value, is_filter = self.reveal(operand, value)

            if is_filter:
                filters.append(self.make_filter(operand, value))
            else:
                res, flt = getattr(self, f'operator_{operator}')(value, operand)
                if flt:
                    filters.append(flt)

                if not res:
                    result = False

        return result, filters

    def make_filter(self, operand: str, value):
        operator = "exact"
        if type(value) is dict:
            operator, value = list(value.items())[0]

        operand = "{}__{}".format(operand.replace(".", "__"), operator)

        return Q(**{operand: value})

    def reveal(self, operand: str, value):
        is_filter = False
        parts = operand.split('.', 1)

        if len(parts) == 2:
            if parts[0] in self.data_source:
                operand = get_attribute_path(self.data_source[parts[0]], parts[1])
            elif parts[0] == 'obj':
                operand = parts[1]
                is_filter = True

        if type(value) is str:
            parts = value.split('.', 1)
            if len(parts) == 2 and parts[0] != 'obj' and parts[0] in self.data_source:
                value = get_attribute_path(self.data_source[parts[0]], parts[1])

        return operand, value, is_filter

    def operator_exact(self, value, operand):
        return operand == value, []

    def operator_not(self, value, operand):
        return operand != value, []

    def operator_in(self, value, operand):
        return operand in value, []

    def operator_lt(self, value, operand):
        return operand < value, []

    def operator_gt(self, value, operand):
        return operand > value, []

    def operator_and(self, conditions: list, *args) -> (bool, list):
        filters = []

        for condition in conditions:
            res, flt = self.evaluate_condition(condition)
            filters.extend(flt)

            if not res:
                return False, []

        return True, filters

    def operator_or(self, conditions: list, *args) -> (bool, list):
        filters = []

        result = False
        for condition in conditions:
            res, flt = self.evaluate_condition(condition)

            if flt:
                filters.extend(flt)

            if res:
                result = True

        return result, reduce(__or__, filters) if filters else []



from rest_framework import serializers

from trood.contrib.django.apps.plugins.models import TroodPluginModel


class TroodPluginSerizalizer(serializers.ModelSerializer):
    class Meta:
        model = TroodPluginModel
        fields = '__all__'



import json

import requests
from django.conf import settings

from trood.core.utils import get_service_token

from trood.contrib.django.auth.engine import TroodABACEngine


class TroodABACMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        rules = {}
        if settings.TROOD_ABAC['RULES_SOURCE'] == 'FILE':
            with open(settings.TROOD_ABAC['RULES_PATH']) as fp:
                rules = json.load(fp)
        elif settings.TROOD_ABAC['RULES_SOURCE'] == 'URL':
            response = requests.get(
                settings.TROOD_ABAC['RULES_PATH'],
                params={"domain": settings.SERVICE_DOMAIN},
                headers={"Authorization": get_service_token()},
            )
            response.raise_for_status()

            rules = response.json().get('data')

        engine = TroodABACEngine(rules)
        setattr(request, "abac", engine)

        response = self.get_response(request)

        return response



import json

from trood.contrib.django.apps.plugins.models import TroodPluginModel


class TroodBasePlugin:
    id = None
    default_config = {}

    @classmethod
    def get_config(cls, key=None):
        config = cls.default_config
        try:
            plugin_config = TroodPluginModel.objects.get(pk=cls.id)
            config = json.loads(plugin_config.config)
        except Exception:
            pass

        if key:
            return config.get(key, None)
        else:
            return config



import os
from setuptools import setup, find_packages


def get_version():
    with open(os.path.join(os.path.curdir, 'VERSION')) as version_file:
        return version_file.read().strip()


setup(
    name='trood',
    version=get_version(),
    packages=find_packages(),
    include_package_data=True,
    author='Trood Inc',
    url='',
    tests_require=[
        'pytest', 'pytest-coverage', 'pyhamcrest', 'teamcity-messages'
    ],
    install_requires=[
        u'requests==2.22.0', 'djangorestframework', 'django', 'pyparsing', 'dateparser', 'django-redis',
    ],
)



import datetime

import pytest
import requests_mock
from hamcrest import assert_that, has_length, equal_to, not_, contains_string,\
     is_, instance_of, is_not

from trood.api.custodian.client import Client
from trood.api.custodian.exceptions import RecordAlreadyExistsException
from trood.api.custodian.records.model import Record


def test_client_retrieves_existing_record(person_record: Record, client: Client):
    record_data = person_record.data
    record_data['id'] = 99
    with requests_mock.Mocker() as mocker:
        mocker.get(
            '/'.join([client.server_url, 'data/{}/{}'.format(person_record.obj, person_record.get_pk())]),
            json={
                'status': 'OK',
                'data': person_record.data
            }
        )
        record = client.records.get(person_record.obj, person_record.get_pk())
        assert_that(record, is_(instance_of(Record)))


def test_client_returns_none_for_nonexistent_record(person_record: Record, client: Client):
    record_data = person_record.data
    record_data['id'] = 99
    with requests_mock.Mocker() as mocker:
        mocker.get(
            '/'.join([client.server_url, 'data/{}/{}'.format(person_record.obj, person_record.get_pk())]),
            json={
                'status': 'FAIL',
                'data': {}
            }
        )
        record = client.records.get(person_record.obj, person_record.get_pk())
        assert_that(record, is_(None))


def test_client_returns_new_record_on_record_creation(person_record: Record, client: Client):
    person_record.id = None
    assert_that(person_record.exists(), is_not(True))
    record_data = person_record.data
    record_data['id'] = 45
    with requests_mock.Mocker() as mocker:
        mocker.post(
            '/'.join([client.server_url, 'data/{}'.format(person_record.obj)]),
            json={
                'status': 'OK',
                'data': record_data
            }
        )
        record = client.records.create(person_record)
        assert_that(record, is_(instance_of(Record)))
        assert_that(record.exists())
        assert_that(record.id, equal_to(45))


def test_client_deletes(person_record: Record, client: Client):
    record_data = person_record.data
    record_data['id'] = 99
    assert_that(person_record.exists())
    record_data = person_record.data
    with requests_mock.Mocker() as mocker:
        mocker.delete(
            '/'.join([client.server_url, 'data/{}'.format(person_record.obj)]),
            json={
                'status': 'OK',
                'data': record_data
            }
        )
        client.records.delete(person_record)
        assert_that(person_record.exists(), is_not(True))


def test_client_returns_updated_record_on_record_update(person_record: Record, client: Client):
    record_data = person_record.data
    record_data['id'] = 99
    record_data['is_active'] = False
    with requests_mock.Mocker() as mocker:
        mocker.patch(
            '/'.join([client.server_url, 'data/{}'.format(person_record.obj)]),
            json={
                'status': 'OK',
                'data': [record_data]
            }
        )
        record = client.records.update(person_record)
        assert_that(record, is_(instance_of(Record)))
        assert_that(record.exists())
        assert_that(record.is_active, is_(False))


def test_client_returns_iterable_of_records_on_bulk_query(person_record: Record, client: Client):
    query = client.records.query('person').filter(address__city_id__eq=45)
    with requests_mock.Mocker() as mocker:
        mocker.get(
            '/'.join([client.server_url, 'data/person']),
            json={
                'status': 'OK',
                'data': [
                    # nothing terrible in is that records are duplicated
                    person_record.data,
                    person_record.data
                ]
            }
        )
        assert_that(len(query), equal_to(2))
        for record in query:
            assert_that(record, instance_of(Record))


def test_client_returns_list_of_records_on_bulk_create(client: Client):
    records_to_create = [
        Record(
            obj='person',
            name='Ivan Petrov',
            is_active=False,
            street="street",
            age=42
        ),
        Record(
            obj='person',
            name='Nikolay Kozlov',
            is_active=True,
            street="street",
            age=36
        )
    ]

    with requests_mock.Mocker() as mocker:
        mocker.post(
            '/'.join([client.server_url, 'data/person']),
            json={
                'status': 'OK',
                'data': [
                    {**x.data, **{'id': records_to_create.index(x) + 1}} for x in records_to_create
                ]
            }
        )
        created_records = client.records.create(*records_to_create)
        for created_record in created_records:
            assert_that(created_record, instance_of(Record))
            assert_that(created_record.name)


def test_client_returns_list_of_records_on_bulk_update(client: Client):
    records_to_update = [
        Record(
            obj='person',
            name='Ivan Petrov',
            is_active=False,
            age=15,
            id=23,
            street=""
        ),
        Record(
            obj='person',
            name='Nikolay Kozlov',
            is_active=True,
            age=44,
            id=34,
            street=""
        )
    ]

    with requests_mock.Mocker() as mocker:
        mocker.patch(
            '/'.join([client.server_url, 'data/person']),
            json={
                'status': 'OK',
                'data': [x.data for x in records_to_update]
            }
        )
        updated_records = client.records.update(*records_to_update)
        for updated_record in updated_records:
            assert_that(updated_record, instance_of(Record))
            assert_that(updated_record.name)
        assert_that(updated_records, equal_to(records_to_update))


def test_client_returns_list_of_records_without_pk_value_on_bulk_delete(client: Client):
    records_to_delete = [
        Record(
            obj='person',
            name='Ivan Petrov',
            is_active=False,
            id=23
        ),
        Record(
            obj='person',
            name='Nikolay Kozlov',
            is_active=True,
            id=34
        )
    ]

    with requests_mock.Mocker() as mocker:
        mocker.delete(
            '/'.join([client.server_url, 'data/person']),
            json={
                'status': 'OK'
            }
        )
        client.records.delete(*records_to_delete)
        for record in records_to_delete:
            assert_that(not record.exists())


class TestCustodianOperationsIntegrationSeries:
    @pytest.mark.integration
    def test_new_record_is_created(self, person_record: Record, client: Client):
        """
        Create a new record and check it has PK field assigned
        :param person_record:
        :param client:
        """
        person_record = client.records.create(person_record)
        assert_that(person_record.id, instance_of(int))
        assert_that(person_record, instance_of(Record))

    @pytest.mark.integration
    def test_record_cration_with_given_id(self, person_record: Record, client: Client):
        """
        Create a new record with given id
        :param person_record:
        :param client:
        """
        person_record.id = 199
        person_record = client.records.create(person_record)
        assert_that(person_record, instance_of(Record))
        assert_that(person_record.id, equal_to(199))


    @pytest.mark.integration
    def test_new_record_can_be_retrieved_by_pk(self, person_record: Record, client: Client):
        """
        Create a new record and check it has PK field assigned
        :param person_record:
        :param client:
        """
        person_record = client.records.create(person_record)
        assert_that(client.records.get(person_record.obj, person_record.get_pk()), instance_of(Record))

    @pytest.mark.integration
    def test_new_record_is_not_created_if_pk_is_duplicated(self, person_record: Record, client: Client):
        """
        Try to create a new record with duplicated PK. Client should raise exception
        :param person_record:
        :param client:
        """
        # create new record
        person_record = client.records.create(person_record)
        with pytest.raises(RecordAlreadyExistsException):
            client.records.create(person_record)

    @pytest.mark.integration
    def test_record_is_updated(self, person_record: Record, client: Client):
        """
        Change the record`s field value and store this change in the database
        :param person_record:
        :param client:
        """
        person_record = client.records.create(person_record)
        new_name = 'Feodor'
        assert_that(new_name, not_(equal_to(person_record.name)))
        person_record.name = new_name
        person_record = client.records.update(person_record)
        assert_that(person_record, instance_of(Record))
        assert_that(person_record.name, equal_to(new_name))

    @pytest.mark.integration
    def test_record_is_deleted(self, person_record: Record, client: Client):
        """
        Delete the record and verify it does not exist in the database
        :param person_record:
        :param client:
        """
        person_record = client.records.create(person_record)
        pk = person_record.get_pk()
        deleted_record = client.records.delete(person_record)
        assert_that(person_record.get_pk(), is_(None))
        assert_that(client.records.get(person_record.obj, pk), is_(None))
        assert_that(deleted_record, instance_of(Record))

    @pytest.mark.integration
    def test_records_are_created(self, client: Client):
        """
        Create two records and verify pk values are assigned
        :param client:
        """
        first_record = Record(obj='person', **{'name': 'Feodor', 'is_active': True, 'age': 23, "street": "St"})
        second_record = Record(obj='person', **{'name': 'Victor', 'is_active': False, 'age': 22, "street": "St"})
        first_record, second_record = client.records.create(first_record, second_record)
        assert_that(first_record.get_pk(), instance_of(int))
        assert_that(second_record.get_pk(), instance_of(int))
        self._created_records = (first_record, second_record)

    @pytest.mark.integration
    def test_records_are_updated(self, two_records, client: Client):
        """
        Update created records and verify updated values are set
        :param client:
        """
        assert_that(two_records[0].name, not_(equal_to('Petr')))
        assert_that(two_records[1].is_active, is_(False))
        two_records[0].name = 'Petr'
        two_records[1].is_active = True
        client.records.update(*two_records)
        assert_that(two_records[0].name, equal_to('Petr'))
        assert_that(two_records[1].is_active, is_(True))

    @pytest.mark.integration
    def test_records_are_deleted(self, two_records, client: Client):
        """
        Delete created records and verify its pk values are None
        :param client:
        """
        deleted_records = client.records.delete(*two_records)
        assert_that(two_records[0].get_pk(), is_(None))
        assert_that(two_records[1].get_pk(), is_(None))
        assert_that(deleted_records, instance_of(list))

    @pytest.mark.integration
    def test_slice(self, client: Client):
        client.records.delete(*[x for x in client.records.query('person')])
        assert_that(client.records.query('person'), has_length(0))
        client.records.create(
            Record(obj='person', **{'name': 'Feodor', 'is_active': True, 'age': 23, 'street': 'street'}),
            Record(obj='person', **{'name': 'Victor', 'is_active': False, 'age': 22, 'street': 'street'}),
            Record(obj='person', **{'name': 'Artem', 'is_active': True, 'age': 35, 'street': 'street'}),
            Record(obj='person', **{'name': 'Anton', 'is_active': False, 'age': 55, 'street': 'street'})
        )
        assert_that(client.records.query('person'), has_length(4))
        two_first_records = client.records.query('person')[:2]
        assert_that(two_first_records, has_length(2))


from trood.api.custodian.command import Command, COMMAND_METHOD
from trood.api.custodian.exceptions import CommandExecutionFailureException, RecordAlreadyExistsException, ObjectUpdateException, \
    RecordUpdateException, CasFailureException, ObjectDeletionException
from trood.api.custodian.records.model import Record
from trood.api.custodian.records.query import Query


class RecordsManager:
    def __init__(self, client):
        self.client = client

    def _get_record_command_name(self, obj_name, record_id=None) -> str:
        """
        Constructs an uri chunk for API communication
        :param obj:
        :param record_id:
        :return:
        """
        args = ['data', obj_name]
        if record_id:
            args.append(str(record_id))
        return '/'.join(args)

    def partial_update(self, obj_name: str, pk, values, **kwargs):
        """
        Performs partial update of existing record
        :param obj_name:        
        """
        data, ok = self.client.execute(
            command=Command(name=self._get_record_command_name(obj_name, pk),
                            method=COMMAND_METHOD.PATCH),
            data=values,
            params=kwargs
        )
        if ok:
            return Record(obj_name, **data)
        else:
            if data.get('code') == 'cas_failed':
                raise CasFailureException(data.get('Msg', ''))
            else:
                raise RecordUpdateException(data.get('Msg', ''))

    def get(self, obj_name: str, record_id: str, **kwargs):
        """
        Retrieves an existing record from Custodian
        :param obj_name:
        :param record_id:
        """
        data, ok = self.client.execute(
            command=Command(name=self._get_record_command_name(obj_name, record_id), method=COMMAND_METHOD.GET),
            params=kwargs
        )
        return Record(obj_name, **data) if ok else None

    def _query(self, obj_name: str, query_string: str, **kwargs):
        """
        Performs an Custodian API call and returns a list of records
        :param obj_name:
        :param query_string:
        """
        if kwargs.get("omit_outers", None) is False:
            del kwargs['omit_outers']
        data, _ = self.client.execute(
            command=Command(name=self._get_record_command_name(obj_name), method=COMMAND_METHOD.GET),
            params={'q': query_string, **kwargs}
        )

        records = []
        for record_data in data:
            records.append(Record(obj_name, **record_data))
        return records

    def query(self, obj_name: str, depth=1, omit_outers=False) -> Query:
        """
        Returns a Query object
        :param obj_name:
        :return:
        """
        return Query(obj_name, self, depth=depth, omit_outers=omit_outers)

    def create(self, *records: Record, **kwargs):
        """
        Creates new records in the Custodian.
        :param records:
        """
        requests_data = self._get_request_data(*records)
        self._check_records_have_same_object(*records)
        obj = records[0].obj
        data, ok = self.client.execute(
            command=Command(name=self._get_record_command_name(obj), method=COMMAND_METHOD.POST),
            data=requests_data, params=kwargs
        )
        records = []
        if ok:
            if isinstance(data, dict):
                return Record(obj, **data)

            elif isinstance(data, list):
                return [Record(obj, **d) for d in data]

        elif data.get('Msg', '').find('duplicate') != -1:
            raise RecordAlreadyExistsException
        else:
            raise CommandExecutionFailureException(data.get('Msg'))

    def update(self, *records: Record, **kwargs):
        """
        Updates new records in the Custodian.
        :param records:
        """
        self._check_records_have_same_object(*records)
        obj = records[0].obj
        data, ok = self.client.execute(
            command=Command(name=self._get_record_command_name(obj), method=COMMAND_METHOD.PATCH),
            data=[record.data for record in records], params=kwargs
        )

        if ok:
            for i, d in enumerate(data):
                print(d)
                records[i].__init__(obj, **d)
            if len(records) == 1:
                return records[0]
            return list(records)
        
        else:
            raise ObjectUpdateException(data.get('Msg'))

    def delete(self, *records: Record):
        """
        Deletes records from the Custodian.
        :param records:
        """
        if records:
            self._check_records_have_same_object(*records)
            obj = records[0].obj
            data, ok = self.client.execute(
                command=Command(name=self._get_record_command_name(obj), method=COMMAND_METHOD.DELETE),
                data=[{"id": record.get_pk()} for record in records]
            )
            if ok:
                for record in records:
                    record.id = None
                if len(records) == 1:
                    return records[0]
                return list(records)
            else:
                raise ObjectDeletionException(data.get('Msg'))
        else:
            return []
            
    def _check_records_have_same_object(self, *records: Record):
        """
        Bulk operations are permitted only for one object at time
        :param records:
        :return:
        """
        obj = records[0].obj
        for record in records[1:]:
            assert obj == record.obj, 'Attempted to perform bulk operation on the list of diverse records'
        return True

    def _get_request_data(self, *records):
        if len(records) == 1:
            return records[0].data
        else:
            return [record.data for record in records]



import re
from functools import reduce
from operator import __or__, __and__, __invert__

from django.conf import settings
from django.core.exceptions import FieldError
from django.db.models import Q, Field, Lookup
from pyparsing import *
from rest_framework.exceptions import ValidationError
from rest_framework.filters import BaseFilterBackend


@Field.register_lookup
class Not(Lookup):
    lookup_name = 'not'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "%s <> %s ESCAPE '\\'" % (lhs, rhs), params


@Field.register_lookup
class Like(Lookup):
    lookup_name = 'like'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        if self.rhs_is_direct_value() and params and not self.bilateral_transforms:
            params[0] = params[0].replace('*', '%')
        return "%s ILIKE %s ESCAPE '\\'" % (lhs, rhs), params


class TroodRQLFilterBackend(BaseFilterBackend):
    """
    Filter that uses a RQL query.

    The RQL query is expected to be passed as a querystring parameter.
    The RQL_FILTER_QUERY_PARAM setting (which defaults to 'rql') specifies the
    name of the querystring parameter used.
    """

    AND = Literal('and').setParseAction(lambda: 'AND')
    OR = Literal('or').setParseAction(lambda: 'OR')
    NOT = Literal('not').setParseAction(lambda: 'NOT')

    EQ = Literal('eq').setParseAction(lambda: 'exact')
    NE = Literal('ne').setParseAction(lambda: 'ne')
    GE = Literal('ge').setParseAction(lambda: 'gte')
    GT = Literal('gt').setParseAction(lambda: 'gt')
    LE = Literal('le').setParseAction(lambda: 'lte')
    LT = Literal('lt').setParseAction(lambda: 'lt')
    IN = Literal('in').setParseAction(lambda: 'in')
    LIKE = Literal('like').setParseAction(lambda: 'like')

    FN = EQ | NE | GE | GT | LE | LT | IN | LIKE

    OB = Literal('(').suppress()
    CB = Literal(')').suppress()
    CM = Literal(',').suppress()

    TRUE = CaselessKeyword('True') + Optional(OB + CB)
    FALSE = CaselessKeyword('False') + Optional(OB + CB)

    BOOL = TRUE.setParseAction(lambda: True) | FALSE.setParseAction(lambda: False)

    NAME = Word(alphas + '_.', alphanums + '_.')
    VALUE = Word(alphanums + ' _.*+-:@/') | Literal('"').suppress() + Word(alphanums + ' _.*@/') + Literal('"').suppress()

    ARRAY = OB + delimitedList(VALUE, ',') + CB
    ARRAY = ARRAY.setParseAction(lambda s, loc, toks: [toks])

    SIMPLE_COND = FN + OB + NAME + CM + (BOOL | VALUE | ARRAY) + CB

    NESTED_CONDS = Forward()
    AGGREGATE = (AND | OR | NOT) + OB + delimitedList(NESTED_CONDS, ',') + CB
    COND = Group(SIMPLE_COND) | Group(AGGREGATE)
    NESTED_CONDS << COND

    QUERY = NESTED_CONDS

    query_param = getattr(settings, 'RQL_FILTER_QUERY_PARAM', 'rql')

    @classmethod
    def parse_rql(cls, rql):
        scan = list(cls.QUERY.scanString(rql))
        if len(scan) > 1:
            part_rql = rql[scan[0][1]:scan[-1][2]]
            rql = rql.replace(part_rql, f'and({part_rql})')
        try:
            parse_results = cls.QUERY.parseString(rql)
        except ParseException:
            return []
        return parse_results.asList()

    @classmethod
    def make_query(cls, data):
        conditions = []
        for fn in data:
            if fn[0] == 'AND':
                res = cls.make_query(fn[1:])
                conditions.append(reduce(__and__, res) if res else [])
            elif fn[0] == 'OR':
                res = cls.make_query(fn[1:])
                conditions.append(reduce(__or__, res) if res else [])
            elif fn[0] == 'NOT':
                res = cls.make_query(fn[1:])
                conditions.append(__invert__(res[0]) if res else [])
            else:
                field = '{}__{}'.format(fn[1].replace('.', '__'), fn[0])
                conditions.append(Q(**{field: convert_numeric(fn[2])}))
        return conditions

    @classmethod
    def get_ordering(cls, data):
        parsed = re.search('sort\(([^\)]+)\)', data)
        if parsed:
            parts = parsed.group(1).replace('+', '').split(',')
            return list(map(lambda a: a.strip(), parts))

        return []

    def filter_queryset(self, request, queryset, view):
        qs = queryset

        if self.query_param in request.GET:
            query_string = ','.join(request.GET.getlist(self.query_param, []))

            if len(query_string):
                condition = self.make_query(self.parse_rql(query_string))

                if not re.match('^(limit|sort)', query_string) and not condition:
                    raise ValidationError('Incorrect rql query')

                try:
                    qs = qs.filter(*condition)
                except (ValueError, FieldError) as error:
                    raise ValidationError(f'Invalid field name or value: {error}')

                ordering = self.get_ordering(query_string)
                qs = qs.order_by(*ordering)

        return qs.distinct()


def convert_numeric(val):
    if type(val) is str:
        if val.isnumeric():
            if '.' in val:
                val = float(val)
            else:
                val = int(val)
    elif type(val) is list:
        for i, a in enumerate(val):
            val[i] = convert_numeric(a)

    return val



class Record:
    __data = None
    obj = None

    def __init__(self, obj: str,  **values):
        if self.__data is None:
            self.__data = {}
        if self.obj is None:
            self.obj = obj
        self.set_data(values)

    def __setattr__(self, key, value):
        if key in self.__class__.__dict__:
            super(Record, self).__setattr__(key, value)
        else:
            self.__data[key] = value

    def __getattr__(self, item):
        return self.__data.get(item)

    @property
    def data(self):
        return self.__data

    def __repr__(self):
        return '<Record #{} of "{}" object>'.format(self.id, self.name)

    def set_data(self, values):
        for key, value in values.items():
            setattr(self, key, value)

    def get_pk(self):
        """
        Returns the record`s primary key value
        """
        return getattr(self, "id", None)

    def exists(self):
        """
        True if the record exists in the Custodian
        :return:
        """
        return self.get_pk() is not None



from django.db import models


class TroodPluginModel(models.Model):
    id = models.CharField(max_length=128, primary_key=True)
    name = models.CharField(max_length=128)
    version = models.CharField(max_length=32)
    active = models.BooleanField(default=False)
    config = models.TextField()



from django.conf.urls import include, url
from rest_framework.routers import DefaultRouter

from trood.contrib.django.apps.plugins import views

router = DefaultRouter()

router.register(r'plugins', views.TroodPluginsViewSet)

urlpatterns = [
    url(r'^api/v1.0/', include((router.urls, 'trood_plugins'), namespace='plugins-api')),
]



from django.conf.urls import include, url
from rest_framework.routers import DefaultRouter

from trood.contrib.django.apps.fixtures import views

router = DefaultRouter()

router.register(r'fixtures', views.TroodFixturesViewSet)

urlpatterns = [
    url(r'^api/v1.0/', include((router.urls, 'trood_fixtures'), namespace='fixtures-api')),
]
